@@ -25,27 +25,36 @@
 
 import socket
 import os
-import re
 import requests
 import json
-from hashlib import sha256
-from urlparse import urljoin
-from urllib import quote
+from urllib.parse import urljoin
+from urllib.parse import quote
 
 import electrum
 from electrum import bitcoin
+from electrum import constants
 from electrum import keystore
 from electrum.bitcoin import *
 from electrum.mnemonic import Mnemonic
 from electrum import version
-from electrum.wallet import Multisig_Wallet, Deterministic_Wallet, Wallet
+from electrum.wallet import Multisig_Wallet, Deterministic_Wallet
 from electrum.i18n import _
-from electrum.plugins import BasePlugin, run_hook, hook
+from electrum.plugins import BasePlugin, hook
 from electrum.util import NotEnoughFunds
+from electrum.storage import STO_EV_USER_PW
 
 # signing_xpub is hardcoded so that the wallet can be restored from seed, without TrustedCoin's server
-signing_xpub = "xpub661MyMwAqRbcGnMkaTx2594P9EDuiEqMq25PM2aeG6UmwzaohgA6uDmNsvSUV8ubqwA3Wpste1hg69XHgjUuCD5HLcEp2QPzyV1HMrPppsL"
-billing_xpub = "xpub6DTBdtBB8qUmH5c77v8qVGVoYk7WjJNpGvutqjLasNG1mbux6KsojaLrYf2sRhXAVU4NaFuHhbD9SvVPRt1MB1MaMooRuhHcAZH1yhQ1qDU"
+def get_signing_xpub():
+    if constants.net.TESTNET:
+        return "tpubD6NzVbkrYhZ4XdmyJQcCPjQfg6RXVUzGFhPjZ7uvRC8JLcS7Hw1i7UTpyhp9grHpak4TyK2hzBJrujDVLXQ6qB5tNpVx9rC6ixijUXadnmY"
+    else:
+        return "xpub661MyMwAqRbcGnMkaTx2594P9EDuiEqMq25PM2aeG6UmwzaohgA6uDmNsvSUV8ubqwA3Wpste1hg69XHgjUuCD5HLcEp2QPzyV1HMrPppsL"
+
+def get_billing_xpub():
+    if constants.net.TESTNET:
+        return "tpubD6NzVbkrYhZ4X11EJFTJujsYbUmVASAYY7gXsEt4sL97AMBdypiH1E9ZVTpdXXEy3Kj9Eqd1UkxdGtvDt5z23DKsh6211CfNJo8bLLyem5r"
+    else:
+        return "xpub6DTBdtBB8qUmH5c77v8qVGVoYk7WjJNpGvutqjLasNG1mbux6KsojaLrYf2sRhXAVU4NaFuHhbD9SvVPRt1MB1MaMooRuhHcAZH1yhQ1qDU"
 
 SEED_PREFIX = version.SEED_PREFIX_2FA
 
@@ -90,11 +99,10 @@ class TrustedCoinCosignerClient(object):
             kwargs['headers']['content-type'] = 'application/json'
         url = urljoin(self.base_url, relative_url)
         if self.debug:
-            print '%s %s %s' % (method, url, data)
+            print('%s %s %s' % (method, url, data))
         response = requests.request(method, url, **kwargs)
         if self.debug:
-            print response.text
-            print
+            print(response.text)
         if response.status_code != 200:
             message = str(response.text)
             if response.headers.get('content-type') == 'application/json':
@@ -192,6 +200,8 @@ server = TrustedCoinCosignerClient(user_agent="Electrum/" + version.ELECTRUM_VER
 
 class Wallet_2fa(Multisig_Wallet):
 
+    wallet_type = '2fa'
+
     def __init__(self, storage):
         self.m, self.n = 2, 3
         Deterministic_Wallet.__init__(self, storage)
@@ -204,42 +214,36 @@ class Wallet_2fa(Multisig_Wallet):
     def get_user_id(self):
         return get_user_id(self.storage)
 
-    def get_max_amount(self, config, inputs, recipient, fee):
-        from electrum.transaction import Transaction
-        sendable = sum(map(lambda x:x['value'], inputs))
-        for i in inputs:
-            self.add_input_info(i)
-        xf = self.extra_fee()
-        _type, addr = recipient
-        if xf and sendable >= xf:
-            billing_address = self.billing_info['billing_address']
-            sendable -= xf
-            outputs = [(_type, addr, sendable),
-                       (TYPE_ADDRESS, billing_address, xf)]
-        else:
-            outputs = [(_type, addr, sendable)]
-        dummy_tx = Transaction.from_io(inputs, outputs)
-        if fee is None:
-            fee = self.estimate_fee(config, dummy_tx.estimated_size())
-        amount = max(0, sendable - fee)
-        return amount, fee
+    def min_prepay(self):
+        return min(self.price_per_tx.keys())
 
-    def extra_fee(self):
+    def num_prepay(self, config):
+        default = self.min_prepay()
+        n = config.get('trustedcoin_prepay', default)
+        if n not in self.price_per_tx:
+            n = default
+        return n
+
+    def extra_fee(self, config):
         if self.can_sign_without_server():
             return 0
+        if self.billing_info is None:
+            self.plugin.start_request_thread(self)
+            return 0
         if self.billing_info.get('tx_remaining'):
             return 0
         if self.is_billing:
             return 0
-        price = int(self.price_per_tx.get(1))
-        assert price <= 100000
+        n = self.num_prepay(config)
+        price = int(self.price_per_tx[n])
+        assert price <= 100000 * n
         return price
 
-    def make_unsigned_transaction(self, coins, outputs, config,
-                                  fixed_fee=None, change_addr=None):
+    def make_unsigned_transaction(self, coins, outputs, config, fixed_fee=None,
+                                  change_addr=None, is_sweep=False):
         mk_tx = lambda o: Multisig_Wallet.make_unsigned_transaction(
             self, coins, o, config, fixed_fee, change_addr)
-        fee = self.extra_fee()
+        fee = self.extra_fee(config) if not is_sweep else 0
         if fee:
             address = self.billing_info['billing_address']
             fee_output = (TYPE_ADDRESS, address, fee)
@@ -271,6 +275,8 @@ class Wallet_2fa(Multisig_Wallet):
             raw_tx = r.get('transaction')
             tx.update(raw_tx)
         self.print_error("twofactor: is complete", tx.is_complete())
+        # reset billing_info
+        self.billing_info = None
 
 
 # Utility functions
@@ -291,7 +297,7 @@ def make_xpub(xpub, s):
 
 def make_billing_address(wallet, num):
     long_id, short_id = wallet.get_user_id()
-    xpub = make_xpub(billing_xpub, long_id)
+    xpub = make_xpub(get_billing_xpub(), long_id)
     version, _, _, _, c, cK = deserialize_xpub(xpub)
     cK, c = bitcoin.CKD_pub(cK, c, num)
     return bitcoin.public_key_to_p2pkh(cK)
@@ -303,6 +309,7 @@ class TrustedCoinPlugin(BasePlugin):
     def __init__(self, parent, config, name):
         BasePlugin.__init__(self, parent, config, name)
         self.wallet_class.plugin = self
+        self.requesting = False
 
     @staticmethod
     def is_valid_seed(seed):
@@ -314,23 +321,43 @@ class TrustedCoinPlugin(BasePlugin):
     def is_enabled(self):
         return True
 
+    def can_user_disable(self):
+        return False
+
     @hook
-    def get_additional_fee(self, wallet, tx):
+    def get_tx_extra_fee(self, wallet, tx):
         if type(wallet) != Wallet_2fa:
             return
+        if wallet.billing_info is None:
+            assert wallet.can_sign_without_server()
+            return None
         address = wallet.billing_info['billing_address']
         for _type, addr, amount in tx.outputs():
             if _type == TYPE_ADDRESS and addr == address:
-                return amount
+                return address, amount
 
     def request_billing_info(self, wallet):
+        if wallet.can_sign_without_server():
+            return
+        self.print_error("request billing info")
         billing_info = server.get(wallet.get_user_id()[1])
         billing_address = make_billing_address(wallet, billing_info['billing_index'])
         assert billing_address == billing_info['billing_address']
         wallet.billing_info = billing_info
         wallet.price_per_tx = dict(billing_info['price_per_tx'])
+        wallet.price_per_tx.pop(1)
+        self.requesting = False
         return True
 
+    def start_request_thread(self, wallet):
+        from threading import Thread
+        if self.requesting is False:
+            self.requesting = True
+            t = Thread(target=self.request_billing_info, args=(wallet,))
+            t.setDaemon(True)
+            t.start()
+            return t
+
     def make_seed(self):
         return Mnemonic('english').make_seed(seed_type='2fa', num_bits=128)
 
@@ -357,17 +384,27 @@ class TrustedCoinPlugin(BasePlugin):
         f = lambda x: wizard.request_passphrase(seed, x)
         wizard.show_seed_dialog(run_next=f, seed_text=seed)
 
+    @classmethod
+    def get_xkeys(self, seed, passphrase, derivation):
+        from electrum.mnemonic import Mnemonic
+        from electrum.keystore import bip32_root, bip32_private_derivation
+        bip32_seed = Mnemonic.mnemonic_to_seed(seed, passphrase)
+        xprv, xpub = bip32_root(bip32_seed, 'standard')
+        xprv, xpub = bip32_private_derivation(xprv, "m/", derivation)
+        return xprv, xpub
+
+    @classmethod
     def xkeys_from_seed(self, seed, passphrase):
         words = seed.split()
         n = len(words)
         # old version use long seed phrases
         if n >= 24:
             assert passphrase == ''
-            xprv1, xpub1 = keystore.xkeys_from_seed(' '.join(words[0:12]), '', "m/")
-            xprv2, xpub2 = keystore.xkeys_from_seed(' '.join(words[12:]), '', "m/")
+            xprv1, xpub1 = self.get_xkeys(' '.join(words[0:12]), '', "m/")
+            xprv2, xpub2 = self.get_xkeys(' '.join(words[12:]), '', "m/")
         elif n==12:
-            xprv1, xpub1 = keystore.xkeys_from_seed(seed, passphrase, "m/0'/")
-            xprv2, xpub2 = keystore.xkeys_from_seed(seed, passphrase, "m/1'/")
+            xprv1, xpub1 = self.get_xkeys(seed, passphrase, "m/0'/")
+            xprv2, xpub2 = self.get_xkeys(seed, passphrase, "m/1'/")
         else:
             raise BaseException('unrecognized seed length')
         return xprv1, xpub1, xprv2, xpub2
@@ -379,20 +416,22 @@ class TrustedCoinPlugin(BasePlugin):
         k2 = keystore.from_xpub(xpub2)
         wizard.request_password(run_next=lambda pw, encrypt: self.on_password(wizard, pw, encrypt, k1, k2))
 
-    def on_password(self, wizard, password, encrypt, k1, k2):
+    def on_password(self, wizard, password, encrypt_storage, k1, k2):
         k1.update_password(None, password)
-        wizard.storage.set_password(password, encrypt)
+        wizard.storage.set_keystore_encryption(bool(password))
+        if encrypt_storage:
+            wizard.storage.set_password(password, enc_version=STO_EV_USER_PW)
         wizard.storage.put('x1/', k1.dump())
         wizard.storage.put('x2/', k2.dump())
         wizard.storage.write()
         msg = [
-            _("Your wallet file is: %s.")%os.path.abspath(wizard.storage.path),
+            _("Your wallet file is: {}.").format(os.path.abspath(wizard.storage.path)),
             _("You need to be online in order to complete the creation of "
               "your wallet.  If you generated your seed on an offline "
-              'computer, click on "%s" to close this window, move your '
+              'computer, click on "{}" to close this window, move your '
               "wallet file to an online computer, and reopen it with "
-              "Electrum.") % _('Cancel'),
-            _('If you are online, click on "%s" to continue.') % _('Next')
+              "Electrum.").format(_('Cancel')),
+            _('If you are online, click on "{}" to continue.').format(_('Next'))
         ]
         msg = '\n\n'.join(msg)
         wizard.stack = []
@@ -429,7 +468,7 @@ class TrustedCoinPlugin(BasePlugin):
         else:
             self.create_keystore(wizard, seed, passphrase)
 
-    def on_restore_pw(self, wizard, seed, passphrase, password, encrypt):
+    def on_restore_pw(self, wizard, seed, passphrase, password, encrypt_storage):
         storage = wizard.storage
         xprv1, xpub1, xprv2, xpub2 = self.xkeys_from_seed(seed, passphrase)
         k1 = keystore.from_xprv(xprv1)
@@ -440,10 +479,14 @@ class TrustedCoinPlugin(BasePlugin):
         storage.put('x1/', k1.dump())
         storage.put('x2/', k2.dump())
         long_user_id, short_id = get_user_id(storage)
-        xpub3 = make_xpub(signing_xpub, long_user_id)
+        xpub3 = make_xpub(get_signing_xpub(), long_user_id)
         k3 = keystore.from_xpub(xpub3)
         storage.put('x3/', k3.dump())
-        storage.set_password(password, encrypt)
+
+        storage.set_keystore_encryption(bool(password))
+        if encrypt_storage:
+            storage.set_password(password, enc_version=STO_EV_USER_PW)
+
         wizard.wallet = Wallet_2fa(storage)
         wizard.create_addresses()
 
@@ -453,7 +496,7 @@ class TrustedCoinPlugin(BasePlugin):
         xpub2 = wizard.storage.get('x2/')['xpub']
         # Generate third key deterministically.
         long_user_id, short_id = get_user_id(wizard.storage)
-        xpub3 = make_xpub(signing_xpub, long_user_id)
+        xpub3 = make_xpub(get_signing_xpub(), long_user_id)
         # secret must be sent by the server
         try:
             r = server.create(xpub1, xpub2, email)
@@ -464,7 +507,8 @@ class TrustedCoinPlugin(BasePlugin):
             if e.status_code == 409:
                 r = None
             else:
-                raise e
+                wizard.show_message(str(e))
+                return
         if r is None:
             otp_secret = None
         else:
@@ -524,9 +568,8 @@ class TrustedCoinPlugin(BasePlugin):
             _, _, _, _, c, k = deserialize_xprv(xprv)
             pk = bip32_private_key([0, 0], k, c)
             key = regenerate_key(pk)
-            compressed = is_compressed(pk)
-            sig = key.sign_message(message, compressed)
-            return base64.b64encode(sig)
+            sig = key.sign_message(message, True)
+            return base64.b64encode(sig).decode()
 
         signatures = [f(x) for x in [xprv1, xprv2]]
         r = server.reset_auth(short_id, challenge, signatures)
